// TestInterface2.cs

using System;

namespace TestInterface {
    interface ICountable {
        bool Tick();
    }

    abstract class ACountable : ICountable {
        public abstract bool Tick();
    }

    class Counter : ICountable {
        public Counter(int c)                    { count = c; }
        public Counter() : this(0)                          { }

        public override string  ToString() { return ""+count; }

        public    int     GetCount()          { return count; }
        protected void    SetCount(int c)        { count = c; }
        public virtual bool    Tick() {
            return ++count == Int32.MaxValue;
        }
        private int count;
    }

    class BoundedCounter : Counter {
        public BoundedCounter(int count, int min, int max,
                              bool directionUp) : base (count) {
            this.min = min;
            this.max = max;
            this.directionUp = directionUp;
            this.increment = 1;
        }
        public BoundedCounter() : this(0, 0, Int32.MaxValue, true) { }
        
        public override string ToString() {
            return GetCount()+"["+GetMin()+".."+GetMax()+"] "+(directionUp?"UP":"DOWN");
        }
        
        // Bumps the count depending of counter's direction and increment. 
        public override bool Tick() {
            if (directionUp) {
                if (GetCount() + GetIncrement()  >  GetMax()) {
                    SetCount(min);
                    return true;
                } else {
                    SetCount(GetCount() + GetIncrement());
                }
            } else {
                if (GetCount() - GetIncrement()  <  GetMin()) {
                    SetCount(max);
                    return true;
                } else {
                    SetCount(GetCount() - GetIncrement());
                }
            }
            return false;
        }
        
        // Gets the minimum count value. 
        public virtual int GetMin() { return min; }

        // Gets the maximum count value. 
        public virtual int GetMax() { return max; }

        // Gets the count's increment. 
        public virtual int GetIncrement() { return increment; }

        // Sets the count's increment. 
        public virtual void SetIncrement(int n) { increment = n; }

        // Gets the count's direction. 
        public virtual bool getDirection() { return directionUp; }

        // Sets the count's direction. 
        public virtual void setDirection(bool value) { 
            if (value != directionUp) directionUp = value;
        }
        private   int      increment;
        private   int      min, max;
        private   bool     directionUp;
    }

    class DownCounter : BoundedCounter {
        public DownCounter(int count) : base(count, 0, 9, false) { }
    }

    class UpCounter : BoundedCounter {
        public UpCounter(int count) : base(count, 0, 59, true) { }
    }

    class CascadedCounter : BoundedCounter {
        public CascadedCounter(int max, ICountable upper) : 
            base(0, 0, max, true) {
            this.upper = upper;
        }
        public override bool Tick() {
            bool overflow = false;

            if ( base.Tick() ) {
                SetCount(GetMin());
                if (upper != null) upper.Tick();
                overflow = true;
            }
            return overflow;
        }
        private ICountable  upper;
    }

    class HourMinute : ICountable {
        public HourMinute() {
            hour   = new CascadedCounter(59, null);
            minute = new CascadedCounter(59, hour);
        }
        public bool Tick() { return minute.Tick(); }
        
        public override string ToString() {
            return String.Format("{0:D2}:{1:D2}", hour.GetCount(), minute.GetCount());
        }
        private Counter hour;
        private Counter minute;
    }

    public class TestInterface {
        public static void Main() {
            Counter up = new BoundedCounter(0, 0, 5, true);
            System.Console.WriteLine("UpCounter [0..5] starting at 0: ");
            for (int n = 0; n < 14; n++) {
                System.Console.Write("{0} ", up.GetCount());
                up.Tick();
            }

            Counter down = new BoundedCounter(5, 0, 5, false);
            System.Console.WriteLine("\n\nDownCounter [5..0] starting at 5: ");
            for (int n = 0; n < 14; n++) {
                System.Console.Write("{0} ", down.GetCount());
                down.Tick();
            }

            Counter ch = new UpCounter(5);
            System.Console.WriteLine("\n\nUpCounter [0..59] starting at 5: ");
            for (int n = 0; n < 61; n++) {
                System.Console.Write("{0} ", ch.GetCount());
                ch.Tick();
            }

            Counter cd = new DownCounter(5);
            System.Console.WriteLine("\n\nDownCounter [9..0] starting at 5: ");
            for (int n = 0; n < 14; n++) {
                System.Console.Write("{0} ", cd.GetCount());
                cd.Tick();
            }

            Counter ch2 = new UpCounter(5);
            System.Console.WriteLine("\n\nUpCounter [0..59] starting at 5 by 2: ");
            ((BoundedCounter)ch2).SetIncrement(2);
            for (int n = 0; n < 61; n++) {
                System.Console.Write("{0} ", ch2.GetCount());
                ch2.Tick();
                
            }

            Counter cd2 = new DownCounter(5);
            System.Console.WriteLine("\n\nDownCounter [9..0] starting at 5 by 2: ");
            ((BoundedCounter)cd2).SetIncrement(2);
            for (int n = 0; n < 14; n++) {
                System.Console.Write("{0} ", cd2.GetCount());
                cd2.Tick();
            }

            System.Console.WriteLine("\n\nHourMinute:");
            HourMinute hhmm = new HourMinute();
            for (int n = 0; n < 64; n++) {
                System.Console.Write("{0} ", hhmm.ToString());
                hhmm.Tick();
            }
        }
    }
}
