using System;
using System.Threading;

class Buffer {
    private readonly int Max;

    public Buffer(int max) {
        Max = max;
        head = tail = count = 0;
        buffer = new char[Max];
        full = false;
        empty = true;
    }
    public void put(char c) {
        lock (this) {        // Critical section access to buffer using lock.
            while (full) Monitor.Wait(this);   // Blocks if buffer full.

            buffer[tail] = c;                  
            if (++count == Max) full = true;
            empty = false;
            tail = (tail + 1) % Max;
            Monitor.Pulse(this);               // Notifies the parallel port.
        }
    }
    public char get() {
        char c;
     // lock (this) {      
        Monitor.Enter(this); // Critical section access to buffer using Monitor.
        try {
            while (empty) Monitor.Wait(this);  // Blocks if buffer empty.

            c = buffer[head];                   
            if (--count == 0) empty = true;
            full = false;
            head = (head + 1) % Max;
            Monitor.PulseAll(this);            // Notifies all serial ports.
        } finally {
            Monitor.Exit(this);
        }    
     // }
        return c;
    }
    private char[] buffer;
    private bool   full, empty;
    private int    head, tail, count;
}

class SerialPortThread {
    public SerialPortThread(Buffer buffer, int num) { 
        this.buffer = buffer;
        this.num = num; 
    }
    private static char inputFromSerial() {
        Thread.Sleep(500); // Waits 0.5 sec (to emulate communication latency).
        return c++;
    }
    public void run() {
        while (true) {
            char c = inputFromSerial();
            Console.WriteLine("sp{0}  {1}", num, c);
            buffer.put(c);
        }
    }
    private static char    c = 'a'; // Shared by all serial ports.
    private        int     num;
    private        Buffer  buffer;
}

class ParallelPortThread {
    public ParallelPortThread(Buffer buffer) {
        this.buffer = buffer;
    }
    public void run() {
        while (true) {
            char c = buffer.get();
            Console.WriteLine("          pp  {0}", c);
            Thread.Sleep(100); // Waits 0.1 sec (to emulate printing latency).
        }
    }
    private Buffer  buffer;
}

class Multiplexer {
    public static void Main() {                   // Main thread entry.
        Buffer             buf = new Buffer(4);        
        SerialPortThread   sp1 = new SerialPortThread(buf, 1);
        SerialPortThread   sp2 = new SerialPortThread(buf, 2);
        ParallelPortThread pp  = new ParallelPortThread(buf);
        
        // Creates all threads passing on their run entry points. 
        Thread s1 = new Thread(new ThreadStart(sp1.run));
        Thread s2 = new Thread(new ThreadStart(sp2.run));
        Thread p  = new Thread(new ThreadStart(pp.run));

        Console.WriteLine("     in       out");

        s1.Start();   // Starts first serial port.
        s2.Start();   // Starts second serial port.
        p.Start();    // Starts parallel port.
    }
}
